package de.be4.smvclipse.prolog;

import java.io.FileWriter;
import java.io.IOException;
import java.io.PrintWriter;
import java.util.HashMap;
import java.util.List;
import java.util.Map;

import de.be4.smvclipse.prolog.exception.IdentifierException;
import de.be4.smvclipse.prolog.exception.TypeViolationException;
import de.be4.smvclipse.sablecc.analysis.DepthFirstAdapter;
import de.be4.smvclipse.sablecc.node.AArrayIdentifier;
import de.be4.smvclipse.sablecc.node.AArrayTypeSpecifier;
import de.be4.smvclipse.sablecc.node.AAsynchronTypeSpecifier;
import de.be4.smvclipse.sablecc.node.ABooleanTypeSpecifier;
import de.be4.smvclipse.sablecc.node.ACaseBodyElement;
import de.be4.smvclipse.sablecc.node.ACaseExpr;
import de.be4.smvclipse.sablecc.node.AComplexAssignConstraint;
import de.be4.smvclipse.sablecc.node.AConditionalExpr;
import de.be4.smvclipse.sablecc.node.AConjunctionExpr;
import de.be4.smvclipse.sablecc.node.ADefineDeclaration;
import de.be4.smvclipse.sablecc.node.ADisjunctionExpr;
import de.be4.smvclipse.sablecc.node.ADivExpr;
import de.be4.smvclipse.sablecc.node.AEnumTypeSpecifier;
import de.be4.smvclipse.sablecc.node.AEqualsExpr;
import de.be4.smvclipse.sablecc.node.AGThanExpr;
import de.be4.smvclipse.sablecc.node.AGeThanExpr;
import de.be4.smvclipse.sablecc.node.AIdentifierExpr;
import de.be4.smvclipse.sablecc.node.AIffExpr;
import de.be4.smvclipse.sablecc.node.AImpliesExpr;
import de.be4.smvclipse.sablecc.node.AInExpr;
import de.be4.smvclipse.sablecc.node.AInitAssignConstraint;
import de.be4.smvclipse.sablecc.node.AIntegerTypeSpecifier;
import de.be4.smvclipse.sablecc.node.AIvarDeclaration;
import de.be4.smvclipse.sablecc.node.ALThanExpr;
import de.be4.smvclipse.sablecc.node.ALeThanExpr;
import de.be4.smvclipse.sablecc.node.ALeftShiftExpr;
import de.be4.smvclipse.sablecc.node.AMinusExpr;
import de.be4.smvclipse.sablecc.node.AModule;
import de.be4.smvclipse.sablecc.node.AModuleParameters;
import de.be4.smvclipse.sablecc.node.AModulesPUnit;
import de.be4.smvclipse.sablecc.node.AModuloExpr;
import de.be4.smvclipse.sablecc.node.AMultExpr;
import de.be4.smvclipse.sablecc.node.ANegationExpr;
import de.be4.smvclipse.sablecc.node.ANegativeIntegerNumber;
import de.be4.smvclipse.sablecc.node.ANequalsExpr;
import de.be4.smvclipse.sablecc.node.ANextAssignConstraint;
import de.be4.smvclipse.sablecc.node.AObjectIdentifier;
import de.be4.smvclipse.sablecc.node.APlusExpr;
import de.be4.smvclipse.sablecc.node.APositiveIntegerNumber;
import de.be4.smvclipse.sablecc.node.ARightShiftExpr;
import de.be4.smvclipse.sablecc.node.ASetExpr;
import de.be4.smvclipse.sablecc.node.ASimpleIdentifier;
import de.be4.smvclipse.sablecc.node.ASubrangeExpr;
import de.be4.smvclipse.sablecc.node.ASynchronTypeSpecifier;
import de.be4.smvclipse.sablecc.node.AUndefinedExpr;
import de.be4.smvclipse.sablecc.node.AUnionExpr;
import de.be4.smvclipse.sablecc.node.AVarDeclaration;
import de.be4.smvclipse.sablecc.node.AXdisjunctionExpr;
import de.be4.smvclipse.sablecc.node.Node;
import de.be4.smvclipse.sablecc.node.PCaseBodyElement;
import de.be4.smvclipse.sablecc.node.PExpr;
import de.be4.smvclipse.sablecc.node.PIdentifier;
import de.be4.smvclipse.sablecc.node.PIntegerNumber;
import de.be4.smvclipse.sablecc.node.Start;
import de.be4.smvclipse.sablecc.node.TAtom;
import de.be4.smvclipse.sablecc.node.TDigit;
import de.be4.smvclipse.sablecc.node.TUndefined;
import de.hhu.stups.sablecc.patch.PositionedNode;
import de.hhu.stups.sablecc.patch.SourcePosition;

public class AST2PrologTransformator extends DepthFirstAdapter{

	private PrologTermOutput out;
	
	private ModuleElementsContainer eltsContainer;
	
	public AST2PrologTransformator(String outputFile) {
		try {
			if ( outputFile.equals("System.out") ) {
				out = new PrologTermOutput(System.out);
			} else {
				out = new PrologTermOutput(new PrintWriter(new FileWriter(outputFile)));
			}
		} catch(IOException ioe) {
			ioe.printStackTrace();
			System.exit(0);
		}
	}
	
	private Map<String, PExpr> definitionsHM = new HashMap<String, PExpr>();
	
	public static String extractField(PIdentifier id) {
		if ( id instanceof ASimpleIdentifier ) {
			return ((ASimpleIdentifier)id).getName().getText();
		} else if ( id instanceof AObjectIdentifier ) {
			return extractField(((AObjectIdentifier)id).getObject());
		} else if ( id instanceof AArrayIdentifier ) {
			return extractField(((AArrayIdentifier)id).getName());
		} else 
			return null;
	}
	
	private void createListAndApplySwitch(List<Node> nodes) {
		out.openList();
		for (Node node : nodes) {
			node.apply(this);
		}
		out.closeList();
	}
	
	@SuppressWarnings("unused")
	private void createEmptyList() {
		out.openList();
		out.closeList();
	}
	
	@SuppressWarnings("unused")
	private void createTermWithEmptyList(String functor_name) {
		out.openTerm(functor_name);
		out.openList();
		out.closeList();
		out.closeTerm();
	}
	
	private int getNumber(PIntegerNumber integer) {
		if ( integer instanceof APositiveIntegerNumber )
			return Integer.parseInt(((APositiveIntegerNumber) integer).getDigit().getText());
		else 
			return Integer.parseInt(((ANegativeIntegerNumber) integer).getDigit().getText());
	}
	
	@Override
	public void outStart(Start node) {
		super.outStart(node);
		
		out.fullstop();
		out.flush();
	}
	
	@SuppressWarnings("unchecked")
	@Override
	public void caseAModulesPUnit(AModulesPUnit node) {
		// smv_unit(modules([..]))
		out.openTerm("smv_unit");
		createListAndApplySwitch((List)node.getModuleList());
		out.closeTerm();	//smv_unit
	}
	
	@SuppressWarnings("unchecked")
	@Override
	public void caseAModule(AModule node) {
		eltsContainer = new ModuleElementsContainer(node.getModuleBodyElements());

		out.openTerm("mod");
		node.getModuleName().apply(this);
		
//		formal parameters
		out.openList();
		
		if ( node.getModuleParameters() != null ) {
			List params = ((AModuleParameters)node.getModuleParameters()).getList();
			for ( int i=0; i<params.size(); i++ ) {
				PIdentifier id = ((AIdentifierExpr)params.get(i)).getIdentifier();
				String fparam = extractField(id);
				AIvarDeclaration ivarDecl = (AIvarDeclaration)eltsContainer.getInputParameterDeclaration(fparam);
				
				if( ivarDecl == null ) {
					continue;
				} else {
					out.openTerm("input", true);
					id.apply(this);
					ivarDecl.getTypeSpecifier().apply(this);
					out.closeTerm();
				}
			}
		}
		
		out.closeList();
		
		createListAndApplySwitch(eltsContainer.getAllVariables());
		createListAndApplySwitch(eltsContainer.getAllDefinitions());
		createListAndApplySwitch(eltsContainer.getAllAssignments());
		
		out.closeTerm(); //mod
	}
	
	@SuppressWarnings("unchecked")
	@Override
	public void caseAModuleParameters(AModuleParameters node) {
//		actual parameters
		createListAndApplySwitch((List)node.getList());		
	}
	
	@Override
	public void caseADefineDeclaration(ADefineDeclaration node) {
		String id = extractField(node.getIdentifier());
		if ( definitionsHM.containsKey(id) ) {
			try {
				throw new IdentifierException("The identifier of definition in line "+ ((PositionedNode)node).getStartPos().getLine() +" is ambiguous.");
			} catch (IdentifierException ie) {
				ie.printStackTrace();
				System.exit(0);
			}
		} else {
			definitionsHM.put(extractField(node.getIdentifier()), node.getExpr());
			new TAtom(id).apply(this);	//this atom does not contain any information about its position
		}
	}
	
	@Override
	public void caseAVarDeclaration(AVarDeclaration node) {
		out.openTerm("var", true);
		node.getIdentifier().apply(this);

		out.printNumber(((PositionedNode)node).getStartPos().getLine());	//line
		out.printNumber(((PositionedNode)node).getStartPos().getPos());		//start pos
		out.printNumber(((PositionedNode)node).getEndPos().getPos());		//end pos
		
		node.getTypeSpecifier().apply(this);
		out.closeTerm();
	}
	
	@Override
	public void caseABooleanTypeSpecifier(ABooleanTypeSpecifier node) {
		out.openTerm("set",true);
		out.openList();
		new TDigit("0").apply(this);
		new TDigit("1").apply(this);
		out.closeList();
		out.closeTerm();	//set
	}
	
	@Override
	public void caseAIntegerTypeSpecifier(AIntegerTypeSpecifier node) {
		out.openTerm("int");
		
		int lowerBound = getNumber(node.getLeftNumber());
		int upperBound = getNumber(node.getRightNumber());
		
		if ( lowerBound > upperBound ) {
			try {
				throw new TypeViolationException();
			} catch(TypeViolationException tve) {
				tve.printStackTrace();
				System.exit(0);
			}
		}
		
		out.openList();
		for ( ; lowerBound <= upperBound; lowerBound++ ) {
			new TDigit(""+lowerBound+"").apply(this);				
		}
		out.closeList();
		
		out.closeTerm();	//int
	}
	
	@SuppressWarnings("unchecked")
	@Override
	public void caseAEnumTypeSpecifier(AEnumTypeSpecifier node) {
		out.openTerm("set");
		createListAndApplySwitch((List)node.getEnumerationTypeValue());
		out.closeTerm();
	}
	
	@Override
	public void caseAArrayTypeSpecifier(AArrayTypeSpecifier node) {
		out.openTerm("array");
		node.getLeftNumber().apply(this);
		node.getRightNumber().apply(this);
		node.getTypeSpecifier().apply(this);
		out.closeTerm();
	}
	
	@SuppressWarnings("unchecked")
	@Override
	public void caseASynchronTypeSpecifier(ASynchronTypeSpecifier node) {
		out.openTerm("module_instance");
		out.printAtom("null");	//indicates a synchronous module instance
		node.getIdentifier().apply(this);
		
		if ( node.getActualParameters() != null ) {
			node.getActualParameters().apply(this);
		}
		else {
			out.openList();
			out.closeList();
		}
		
		out.closeTerm();	//module_instance
	}
	
	@SuppressWarnings("unchecked")
	@Override
	public void caseAAsynchronTypeSpecifier(AAsynchronTypeSpecifier node) {
		out.openTerm("module_instance");
		out.printAtom("process");
		node.getIdentifier().apply(this);
		
		node.getActualParameters().apply(this);
		
		out.closeTerm();	//module_instance
	}
	
	@Override
	public void caseAInitAssignConstraint(AInitAssignConstraint node) {
		out.openTerm("init", true);
		node.getIdentifier().apply(this);

		out.printNumber(((PositionedNode)node).getStartPos().getLine());	//line
		out.printNumber(((PositionedNode)node).getStartPos().getPos());	//start pos
		out.printNumber(((PositionedNode)node).getEndPos().getPos());	//end pos

		out.closeTerm();	//init
		
		node.getExpr().apply(this);
	}
	
	@Override
	public void caseANextAssignConstraint(ANextAssignConstraint node) {
		out.openTerm("next", true);
		node.getIdentifier().apply(this);

		out.printNumber(((PositionedNode)node).getStartPos().getLine());	//line
		out.printNumber(((PositionedNode)node).getStartPos().getPos());	//start pos
		out.printNumber(((PositionedNode)node).getEndPos().getPos());	//end pos

		out.closeTerm();	//next
		
		node.getExpr().apply(this);
	}
	
	@Override
	public void caseAComplexAssignConstraint(AComplexAssignConstraint node) {
		AInitAssignConstraint initConstraint = new AInitAssignConstraint((PIdentifier)node.getIdentifier().clone(), (PExpr)node.getExpr().clone());
		((PositionedNode)initConstraint).setStartPos(((PositionedNode)node).getStartPos());
		((PositionedNode)initConstraint).setEndPos(((PositionedNode)node).getEndPos());
		
		ANextAssignConstraint nextConstraint = new ANextAssignConstraint((PIdentifier)node.getIdentifier().clone(), (PExpr)node.getExpr().clone());
		((PositionedNode)nextConstraint).setStartPos(((PositionedNode)node).getStartPos());
		((PositionedNode)nextConstraint).setEndPos(((PositionedNode)node).getEndPos());
		
		initConstraint.apply(this);
		nextConstraint.apply(this);
		
//		out.printNumber(((PositionedNode)node).getStartPos().getLine());	//line
//		out.printNumber(((PositionedNode)node).getStartPos().getPos());	//start pos
//		out.printNumber(((PositionedNode)node).getEndPos().getPos());	//end pos
//		
//		node.getExpr().apply(this);
	}
	
	@Override
	public void caseASubrangeExpr(ASubrangeExpr node) {
		out.openTerm("..");
		node.getLeft().apply(this);
		node.getRight().apply(this);
		out.closeTerm();
	}
	
	@Override
	public void caseAConditionalExpr(AConditionalExpr node) {
		// if (expr, expr, expr)
		out.openTerm("if");

		node.getCondition().apply(this);
		node.getContidionTrueExpr().apply(this);
		node.getConditionFalseExpr().apply(this);
		
		out.closeTerm();
	}
	
	@Override
	public void caseAImpliesExpr(AImpliesExpr node) {
		out.openTerm("->", true);
		node.getLeft().apply(this);
		node.getRight().apply(this);
		out.closeTerm();
	}
	
	@Override
	public void caseAIffExpr(AIffExpr node) {
		out.openTerm("<=>", true);
		node.getLeft().apply(this);
		node.getRight().apply(this);
		out.closeTerm();
	}
	
	@Override
	public void caseADisjunctionExpr(ADisjunctionExpr node) {
		out.openTerm("|", true);
		node.getLeft().apply(this);
		node.getRight().apply(this);
		out.closeTerm();
	}
	
	@Override
	public void caseAXdisjunctionExpr(AXdisjunctionExpr node) {
		out.openTerm("^", true);
		node.getLeft().apply(this);
		node.getRight().apply(this);
		out.closeTerm();
	}
	
	@Override
	public void caseAConjunctionExpr(AConjunctionExpr node) {
		out.openTerm("&", true);
		node.getLeft().apply(this);
		node.getRight().apply(this);
		out.closeTerm();
	}
	
	@Override
	public void caseANegationExpr(ANegationExpr node) {
		out.openTerm("not", true);
		node.getExpr().apply(this);
		out.closeTerm();
	}
	
	@Override
	public void caseAEqualsExpr(AEqualsExpr node) {
		out.openTerm("=", true);
		node.getLeft().apply(this);
		node.getRight().apply(this);
		out.closeTerm();
	}
	
	@Override
	public void caseANequalsExpr(ANequalsExpr node) {
		out.openTerm("~=", true);
		node.getLeft().apply(this);
		node.getRight().apply(this);
		out.closeTerm();
	}
	
	@Override
	public void caseALThanExpr(ALThanExpr node) {
		out.openTerm("<", true);
		node.getLeft().apply(this);
		node.getRight().apply(this);
		out.closeTerm();
	}
	
	@Override
	public void caseAGThanExpr(AGThanExpr node) {
		out.openTerm(">", true);
		node.getLeft().apply(this);
		node.getRight().apply(this);
		out.closeTerm();
	}
	
	@Override
	public void caseALeThanExpr(ALeThanExpr node) {
		out.openTerm("<=", true);
		node.getLeft().apply(this);
		node.getRight().apply(this);
		out.closeTerm();
	}
	
	@Override
	public void caseAGeThanExpr(AGeThanExpr node) {
		out.openTerm(">=", true);
		node.getLeft().apply(this);
		node.getRight().apply(this);
		out.closeTerm();
	}
	
	@Override
	public void caseAUnionExpr(AUnionExpr node) {
		out.openTerm("union", true);
		node.getLeft().apply(this);
		node.getRight().apply(this);
		out.closeTerm();
	}
	
	@Override
	public void caseAInExpr(AInExpr node) {
		out.openTerm("inclusion", true);
		node.getLeft().apply(this);
		node.getRight().apply(this);
		out.closeTerm();
	}
	
	@Override
	public void caseAModuloExpr(AModuloExpr node) {
		out.openTerm("mod", true);
		node.getLeft().apply(this);
		node.getRight().apply(this);
		out.closeTerm();
	}
	
	@Override
	public void caseAPlusExpr(APlusExpr node) {
		out.openTerm("+", true);
		node.getLeft().apply(this);
		node.getRight().apply(this);
		out.closeTerm();
	}
	
	@Override
	public void caseAMinusExpr(AMinusExpr node) {
		out.openTerm("-", true);
		node.getLeft().apply(this);
		node.getRight().apply(this);
		out.closeTerm();
	}
	
	@Override
	public void caseAMultExpr(AMultExpr node) {
		out.openTerm("*", true);
		node.getLeft().apply(this);
		node.getRight().apply(this);
		out.closeTerm();
	}
	
	@Override
	public void caseADivExpr(ADivExpr node) {
		out.openTerm("/", true);
		node.getLeft().apply(this);
		node.getRight().apply(this);
		out.closeTerm();
	}
	
	@Override
	public void caseALeftShiftExpr(ALeftShiftExpr node) {
		out.openTerm("<<", true);
		node.getLeft().apply(this);
		node.getRight().apply(this);
		out.closeTerm();
	}
	
	@Override
	public void caseARightShiftExpr(ARightShiftExpr node) {
		out.openTerm(">>", true);
		node.getLeft().apply(this);
		node.getRight().apply(this);
		out.closeTerm();
	}
	
	@SuppressWarnings("unchecked")
	@Override
	public void caseASetExpr(ASetExpr node) {
		out.openTerm("set");
		createListAndApplySwitch((List)node.getExpr());
		out.closeTerm();
	}
	
	@Override
	public void caseAIdentifierExpr(AIdentifierExpr node) {
		node.getIdentifier().apply(this);
	}
	
	@SuppressWarnings("unchecked")
	@Override
	public void caseACaseExpr(ACaseExpr node) {
		/*
		 * A case-expression can be rewritten as a list of short conditional expressions. For instance
		 * 
		 * next(x) := case {
		 * 	z=0	:	init(x) := 1;
		 * 	z=1 :	init(x) := 0;
		 * }
		 * 
		 * is exactly equivalent to:
		 * 
		 * next(x) := 
		 * 	if (z=0) {
		 * 		init(x) := 1;
		 * 	} else {
		 * 		if (z=1) {
		 * 			init(x) := 0;
		 * 		} else {
		 * 			undefined;
		 * 		}
		 *	}
		 * 
		 * so we do this here.
		 */
		
		List<PCaseBodyElement> caseElts = (List)node.getCaseBodyElement();

		ACaseBodyElement caseElt = (ACaseBodyElement)caseElts.get(0);
		AConditionalExpr condExpr = buildConditionalExpr(caseElt.getCondition(), caseElt.getExpr(), 0, caseElts);
		
		node.replaceBy(condExpr);
		condExpr.apply(this);
	}
	
	private AConditionalExpr buildConditionalExpr(PExpr cond_expr, PExpr ctrue_expr, int curr, List<PCaseBodyElement> caseExprBodies) {
		AConditionalExpr expr = new AConditionalExpr();
		
		expr.setCondition(cond_expr);
		expr.setContidionTrueExpr(ctrue_expr);
		
		curr++;
		if ( curr != caseExprBodies.size() ) {
			ACaseBodyElement caseB = (ACaseBodyElement)caseExprBodies.get(curr);
			
			AConditionalExpr next_expr = buildConditionalExpr(caseB.getCondition(), caseB.getExpr(), curr, caseExprBodies);
			expr.setConditionFalseExpr(next_expr);
		} else {
			expr.setConditionFalseExpr(new AUndefinedExpr(new TUndefined()));
		}
		
		return expr;
	}
	
	@Override
	public void caseASimpleIdentifier(ASimpleIdentifier node) {
		if ( definitionsHM.containsKey(node.getName().getText()) ) {
			definitionsHM.get(node.getName().getText()).apply(this);
		} else {
			node.getName().apply(this);
		}
	}
	
	@Override
	public void caseAObjectIdentifier(AObjectIdentifier node) {
		out.openTerm("obj", true);
		node.getObject().apply(this);
		node.getName().apply(this);
		out.closeTerm();
	}
	
	@Override
	public void caseTAtom(TAtom node) {
		out.openTerm("id", true);
		out.printAtom(node.getText());
		out.printNumber(node.getLine());
		out.printNumber(node.getPos());
		out.printNumber(node.getPos()+node.getText().length());
		out.closeTerm();
	}
	
	@Override
	public void caseTDigit(TDigit node) {
		out.openTerm("num");
		out.printNumber(Integer.parseInt(node.getText()));
		out.closeTerm();
	}
	
}
